@@ -15,42 +15,44 @@
  *
  * The Initial Developer of the Original Code is
  * Wladimir Palant.
- * Portions created by the Initial Developer are Copyright (C) 2006-2010
+ * Portions created by the Initial Developer are Copyright (C) 2006-2011
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s):
  *
  * ***** END LICENSE BLOCK ***** */
 
-var effectiveTLD = Cc["@mozilla.org/network/effective-tld-service;1"].getService(Ci.nsIEffectiveTLDService);
-var RequestList = abp.RequestList;
+Cu.import("resource://gre/modules/XPCOMUtils.jsm");
 
 // Main browser window
 var mainWin = parent;
 
 // The window handler currently in use
-var wndData = null;
+var requestNotifier = null;
 
 var cacheSession = null;
 var noFlash = false;
 
-// Matchers for disabled filters
-var disabledBlacklistMatcher = new abp.Matcher();
-var disabledWhitelistMatcher = new abp.Matcher();
+// Matcher for disabled filters
+var disabledMatcher = new CombinedMatcher();
+
+// Cached string values
+var docDomainThirdParty = null;
+var docDomainFirstParty = null;
 
 var abpHooks = null;
 
-/**
- * Cached value of "enabled" preference, for onPrefChange.
- */
-var oldEnabled = abp.prefs.enabled;
+let lastSelectionProp = "abpSelected" + RequestNotifier.getDataSeed();
 
 function init() {
+  docDomainThirdParty = document.documentElement.getAttribute("docDomainThirdParty");
+  docDomainFirstParty = document.documentElement.getAttribute("docDomainFirstParty");
+
   var list = E("list");
   list.view = treeView;
 
   // Restore previous state
-  var params = abp.getParams();
+  var params = Utils.getParams();
   if (params && params.filter)
   {
     E("searchField").value = params.filter;
@@ -88,29 +90,19 @@ function init() {
   abpHooks = mainWin.document.getElementById("abp-hooks");
   window.__defineGetter__("content", function() {return abpHooks.getBrowser().contentWindow;});
 
-  // Install item listener
-  RequestList.addListener(handleItemChange);
-
-  // Initialize matchers for disabled filters
+  // Initialize matcher for disabled filters
   reloadDisabledFilters();
-  filterStorage.addFilterObserver(reloadDisabledFilters);
-  filterStorage.addSubscriptionObserver(reloadDisabledFilters);
-  abp.prefs.addListener(onPrefChange);
+  FilterStorage.addObserver(reloadDisabledFilters);
+  Prefs.addListener(onPrefChange);
 
   // Activate flasher
   list.addEventListener("select", onSelectionChange, false);
 
-  // Retrieve data for the window
-  wndData = RequestList.getDataForWindow(window.content);
-  treeView.setData(wndData.getAllLocations());
-  if (wndData.lastSelection) {
-    noFlash = true;
-    treeView.selectItem(wndData.lastSelection);
-    noFlash = false;
-  }
+  // Initialize data
+  handleLocationChange();
 
-  // Install a handler for tab changes
-  abpHooks.getBrowser().addEventListener("select", handleTabChange, false);
+  // Install a progress listener to catch location changes
+  abpHooks.getBrowser().addProgressListener(progressListener);
 }
 
 // To be called for a detached window when the main window has been closed
@@ -120,16 +112,12 @@ function mainUnload() {
 
 // To be called on unload
 function cleanUp() {
-  if (!abp)
-    return;
-
   flasher.stop();
-  RequestList.removeListener(handleItemChange);
-  filterStorage.removeFilterObserver(reloadDisabledFilters);
-  filterStorage.removeSubscriptionObserver(reloadDisabledFilters);
-  abp.prefs.removeListener(onPrefChange);
+  requestNotifier.shutdown();
+  FilterStorage.removeObserver(reloadDisabledFilters);
+  Prefs.removeListener(onPrefChange);
 
-  abpHooks.getBrowser().removeEventListener("select", handleTabChange, false);
+  abpHooks.getBrowser().removeProgressListener(progressListener);
   mainWin.removeEventListener("unload", mainUnload, false);
 }
 
@@ -137,38 +125,34 @@ function cleanUp() {
  * Tracks preference changes, calls reloadDisabledFilters whenever Adblock Plus
  * is enabled/disabled.
  */
-function onPrefChange()
+function onPrefChange(name)
 {
-  if (abp.prefs.enabled != oldEnabled)
-  {
-    oldEnabled = abp.prefs.enabled;
+  if (name == "enabled")
     reloadDisabledFilters();
-  }
 }
 
 /**
- * Updates matchers for disabled filters (global disabledBlacklistMatcher and
- * disabledWhitelistMatcher variables), called on each filter change.
+ * Updates matcher for disabled filters (global disabledMatcher variable),
+ * called on each filter change.
  */
 function reloadDisabledFilters()
 {
-  disabledBlacklistMatcher.clear();
-  disabledWhitelistMatcher.clear();
+  disabledMatcher.clear();
 
-  if (abp.prefs.enabled)
+  if (Prefs.enabled)
   {
-    for each (let subscription in filterStorage.subscriptions)
+    for each (let subscription in FilterStorage.subscriptions)
     {
       if (subscription.disabled)
         continue;
   
       for each (let filter in subscription.filters)
-        if (filter instanceof abp.RegExpFilter && filter.disabled)
-          (filter instanceof abp.BlockingFilter ? disabledBlacklistMatcher : disabledWhitelistMatcher).add(filter);
+        if (filter instanceof RegExpFilter && filter.disabled)
+          disabledMatcher.add(filter);
     }
   }
 
-  treeView.setData(treeView.allData);
+  treeView.updateFilters();
 }
 
 // Called whenever list selection changes - triggers flasher
@@ -178,52 +162,30 @@ function onSelectionChange() {
     E("copy-command").removeAttribute("disabled");
   else
     E("copy-command").setAttribute("disabled", "true");
-  if (item && wndData)
-    wndData.lastSelection = item;
+
+  if (item && window.content)
+  {
+    let key = item.location + " " + item.type + " " + item.docDomain;
+    window.content.document.setUserData(lastSelectionProp, key, null);
+    treeView.itemToSelect = null;
+  }
 
   if (!noFlash)
     flasher.flash(item ? item.nodes : null);
 }
 
-function handleItemChange(wnd, type, data, item) {
-  // Check whether this applies to us
-  if (wnd != window.content)
-    return;
+function handleLocationChange()
+{
+  if (requestNotifier)
+    requestNotifier.shutdown();
 
-  // Maybe we got called twice
-  if (type == "select" && data == wndData)
-    return;
-
-  // If adding something from a new data container - select it
-  if (type == "add" && data != wndData)
-    type = "select";
-
-  var i;
-  var filterSuggestions = E("suggestionsList");
-  if (type == "clear") {
-    // Current document has been unloaded, clear list
-    wndData = null;
-    treeView.setData([]);
-  }
-  else if (type == "select" || type == "refresh") {
-    // We moved to a different document, reload list
-    wndData = data;
-    treeView.setData(wndData.getAllLocations());
-  }
-  else if (type == "invalidate")
-    treeView.boxObject.invalidate();
-  else if (type == "add")
-    treeView.addItem(item);
-}
-
-function handleTabChange() {
-  wndData = RequestList.getDataForWindow(window.content);
-  treeView.setData(wndData.getAllLocations());
-  if (wndData.lastSelection) {
-    noFlash = true;
-    treeView.selectItem(wndData.lastSelection);
-    noFlash = false;
-  }
+  treeView.clearData();
+  treeView.itemToSelect = window.content.document.getUserData(lastSelectionProp);
+  requestNotifier = new RequestNotifier(window.content, function(wnd, node, item, scanComplete)
+  {
+    if (item)
+      treeView.addItem(node, item, scanComplete);
+  });
 }
 
 // Fills a box with text splitting it up into multiple lines if necessary
@@ -252,7 +214,7 @@ function fillInTooltip(e) {
   if (!item)
     return false;
 
-  let filter = ("filter" in item && item.filter && !item.filter.disabled ? item.filter : null);
+  let filter = ("filter" in item && item.filter ? item.filter : null);
   let size = ("tooltip" in item ? null : getItemSize(item));
   let subscriptions = (filter ? filter.subscriptions.filter(function(subscription) { return !subscription.disabled; }) : []);
 
@@ -272,7 +234,7 @@ function fillInTooltip(e) {
     setMultilineContent(E("tooltipAddress"), item.location);
   
     var type = item.localizedDescr;
-    if (filter && filter instanceof abp.WhitelistFilter)
+    if (filter && filter instanceof WhitelistFilter)
       type += " " + E("tooltipType").getAttribute("whitelisted");
     else if (filter && item.typeDescr != "ELEMHIDE")
       type += " " + E("tooltipType").getAttribute("filtered");
@@ -281,25 +243,34 @@ function fillInTooltip(e) {
     if (size)
       E("tooltipSize").setAttribute("value", size.join(" x "));
 
-    E("tooltipDocDomain").setAttribute("value", item.docDomain);
+    E("tooltipDocDomain").setAttribute("value", item.docDomain + " " + (item.thirdParty ? docDomainThirdParty : docDomainFirstParty));
   }
 
   if (filter)
   {
-    setMultilineContent(E("tooltipFilter"), filter.text);
+    let filterField = E("tooltipFilter");
+    setMultilineContent(filterField, filter.text);
+    if (filter.disabled)
+    {
+      let disabledText = document.createElement("description");
+      disabledText.className = "disabledTextLabel";
+      disabledText.textContent = filterField.getAttribute("disabledText");
+      filterField.appendChild(disabledText);
+    }
+
     if (subscriptions.length)
     {
       let sourceElement = E("tooltipFilterSource");
       while (sourceElement.firstChild)
         sourceElement.removeChild(sourceElement.firstChild);
-      for each (let subscription in subscriptions)
-        setMultilineContent(sourceElement, subscription.title, true);
+      for (let i = 0; i < subscriptions.length; i++)
+        setMultilineContent(sourceElement, subscriptions[i].title, true);
     }
   }
 
-  var showPreview = prefs.previewimages && !("tooltip" in item);
-  showPreview = showPreview && (item.typeDescr == "IMAGE" || item.typeDescr == "BACKGROUND");
-  showPreview = showPreview && (!item.filter || item.filter instanceof abp.WhitelistFilter);
+  var showPreview = Prefs.previewimages && !("tooltip" in item);
+  showPreview = showPreview && item.typeDescr == "IMAGE";
+  showPreview = showPreview && (!item.filter || item.filter instanceof WhitelistFilter);
   if (showPreview) {
     // Check whether image is in cache (stolen from ImgLikeOpera)
     if (!cacheSession) {
@@ -364,12 +335,12 @@ function fillInContext(/**Event*/ e)
     menuItem.setAttribute("label", menuItem.getAttribute("labeltempl").replace(/\?1\?/, filter.text));
     menuItem.hidden = false;
 
-    if (filter instanceof abp.ActiveFilter && !filter.disabled && filter.subscriptions.length && !filter.subscriptions.some(function(subscription) !(subscription instanceof abp.SpecialSubscription)))
+    if (filter instanceof ActiveFilter && !filter.disabled && filter.subscriptions.length && !filter.subscriptions.some(function(subscription) !(subscription instanceof SpecialSubscription)))
     {
       let domain = null;
       try {
         domain = content.location.host;
-        domain = effectiveTLD.getBaseDomainFromHost(domain);
+        domain = Utils.effectiveTLD.getBaseDomainFromHost(domain);
       } catch (e) {}
 
       if (domain && !filter.isActiveOnlyOnDomain(domain))
@@ -384,12 +355,12 @@ function fillInContext(/**Event*/ e)
     }
   }
 
-  E("contextWhitelist").hidden = ("tooltip" in item || !item.filter || item.filter.disabled || item.filter instanceof abp.WhitelistFilter || item.typeDescr == "ELEMHIDE");
+  E("contextWhitelist").hidden = ("tooltip" in item || !item.filter || item.filter.disabled || item.filter instanceof WhitelistFilter || item.typeDescr == "ELEMHIDE");
   E("contextBlock").hidden = !E("contextWhitelist").hidden;
   E("contextBlock").setAttribute("disabled", "filter" in item && item.filter && !item.filter.disabled);
   E("contextEditFilter").setAttribute("disabled", !("filter" in item && item.filter));
   E("contextOpen").setAttribute("disabled", "tooltip" in item || item.typeDescr == "ELEMHIDE");
-  E("contextFlash").setAttribute("disabled", "tooltip" in item || !(item.typeDescr in visual) || (item.filter && !item.filter.disabled && !(item.filter instanceof abp.WhitelistFilter)));
+  E("contextFlash").setAttribute("disabled", "tooltip" in item || !(item.typeDescr in visual) || (item.filter && !item.filter.disabled && !(item.filter instanceof WhitelistFilter)));
   E("contextCopyFilter").setAttribute("disabled", !allItems.some(function(item) {return "filter" in item && item.filter}));
 
   return true;
@@ -432,36 +403,30 @@ function handleDblClick(event)
  */
 function openInTab(item, /**Event*/ event)
 {
-  if (!item)
-    item = treeView.getSelectedItem();
-  if (!item || item.typeDescr == "ELEMHIDE")
-    return;
-
-  abp.loadInBrowser(item.location, mainWin, event);
+  let items = (item ? [item] : treeView.getAllSelectedItems());
+  for each (let item in items)
+  {
+    if (item && item.typeDescr != "ELEMHIDE")
+      Utils.loadInBrowser(item.location, mainWin, event);
+  }
 }
 
 function doBlock() {
-  if (!abp)
-    return;
-
   var item = treeView.getSelectedItem();
   if (!item || item.typeDescr == "ELEMHIDE")
     return;
 
   var filter = null;
-  if ("filter" in item && item.filter && !item.filter.disabled)
+  if (item.filter && !item.filter.disabled)
     filter = item.filter;
 
-  if (filter && filter instanceof abp.WhitelistFilter)
+  if (filter && filter instanceof WhitelistFilter)
     return;
 
-  openDialog("chrome://adblockplus/content/ui/composer.xul", "_blank", "chrome,centerscreen,resizable,dialog=no,dependent", window.content, item);
+  openDialog("chrome://adblockplus/content/ui/composer.xul", "_blank", "chrome,centerscreen,resizable,dialog=no,dependent", item.nodes, item.orig);
 }
 
 function editFilter() {
-  if (!abp)
-    return;
-
   var item = treeView.getSelectedItem();
   if (treeView.data && !treeView.data.length)
     item = treeView.getDummyTooltip();
@@ -472,16 +437,13 @@ function editFilter() {
   if (!("location") in item)
     item.location = undefined
 
-  abp.openSettingsDialog(item.location, item.filter);
+  Utils.openSettingsDialog(item.location, item.filter);
 }
 
 function enableFilter(filter, enable) {
-  if (!abp)
-    return;
-
   filter.disabled = !enable;
-  filterStorage.triggerFilterObservers(enable ? "enable" : "disable", [filter]);
-  filterStorage.saveToDisk();
+  FilterStorage.triggerObservers(enable ? "filters enable" : "filters disable", [filter]);
+  FilterStorage.saveToDisk();
 
   treeView.boxObject.invalidate();
 }
@@ -494,9 +456,9 @@ function disableOnSite(item, /**Filter*/ filter, /**String*/ domain)
   // Generate text for new filter that excludes current domain
   domain = domain.toUpperCase();
   let text = filter.text;
-  if (filter instanceof abp.RegExpFilter)
+  if (filter instanceof RegExpFilter)
   {
-    if (abp.Filter.optionsRegExp.test(text))
+    if (Filter.optionsRegExp.test(text))
     {
       let found = false;
       let options = RegExp.$1.toUpperCase().split(",");
@@ -514,12 +476,12 @@ function disableOnSite(item, /**Filter*/ filter, /**String*/ domain)
       if (!found)
         options.push("DOMAIN=~" + domain);
 
-      text = text.replace(abp.Filter.optionsRegExp, "$" + options.join(",").toLowerCase());
+      text = text.replace(Filter.optionsRegExp, "$" + options.join(",").toLowerCase());
     }
     else
       text += "$domain=~" + domain.toLowerCase();
   }
-  else if (filter instanceof abp.ElemHideFilter)
+  else if (filter instanceof ElemHideFilter)
   {
     if (/^([^#]+)(#.*)/.test(text))
     {
@@ -536,19 +498,19 @@ function disableOnSite(item, /**Filter*/ filter, /**String*/ domain)
     return;   // Just in case, shouldn't happen
 
   // Insert new filter before the old one and remove the old one then
-  let newFilter = abp.Filter.fromText(text);
+  let newFilter = Filter.fromText(text);
   if (newFilter.disabled && newFilter.subscriptions.length)
   {
     newFilter.disabled = false;
-    filterStorage.triggerFilterObservers("enable", [newFilter]);
+    FilterStorage.triggerObservers("filters enable", [newFilter]);
   }
   else if (!newFilter.subscriptions.length)
   {
     newFilter.disabled = false;
-    filterStorage.addFilter(newFilter, filter);
+    FilterStorage.addFilter(newFilter, filter);
   }
-  filterStorage.removeFilter(filter);
-  filterStorage.saveToDisk();
+  FilterStorage.removeFilter(filter);
+  FilterStorage.saveToDisk();
 
   // Update display
   item.filter = null;
@@ -556,21 +518,15 @@ function disableOnSite(item, /**Filter*/ filter, /**String*/ domain)
 }
 
 function copyToClipboard() {
-  if (!abp)
-    return;
-
   var items = treeView.getAllSelectedItems();
   if (!items.length)
     return;
 
   var clipboardHelper = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper);
-  clipboardHelper.copyString(items.map(function(item) {return item.location}).join(abp.getLineBreak()));
+  clipboardHelper.copyString(items.map(function(item) {return item.location}).join(Utils.getLineBreak()));
 }
 
 function copyFilter() {
-  if (!abp)
-    return;
-
   var items = treeView.getAllSelectedItems().filter(function(item) {return item.filter});
   if (treeView.data && !treeView.data.length)
     items = [treeView.getDummyTooltip()];
@@ -579,13 +535,10 @@ function copyFilter() {
     return;
 
   var clipboardHelper = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper);
-  clipboardHelper.copyString(items.map(function(item) {return item.filter.text}).join(abp.getLineBreak()));
+  clipboardHelper.copyString(items.map(function(item) {return item.filter.text}).join(Utils.getLineBreak()));
 }
 
 function selectAll() {
-  if (!abp)
-    return;
-
   treeView.selectAll();
 }
 
@@ -604,7 +557,7 @@ function saveState() {
     focus: (focused ? focused.id : null),
     position: position
   };
-  abp.setParams(params);
+  Utils.setParams(params);
 }
 
 // closes the sidebar
@@ -616,28 +569,22 @@ function doClose()
 // detaches/reattaches the sidebar
 function detach(doDetach)
 {
-  if (!abp)
-    return;
-
   saveState();
 
   // Store variables locally, global variables will go away when we are closed
-  let myPrefs = prefs;
+  let myPrefs = Prefs;
   let myMainWin = mainWin;
 
   // Close sidebar and open detached window
   myMainWin.document.getElementById("abp-command-sidebar").doCommand();
   myPrefs.detachsidebar = doDetach;
   myMainWin.document.getElementById("abp-command-sidebar").doCommand();
-
-  // Save setting
-  myPrefs.save();
 }
 
 // Returns items size in the document if available
 function getItemSize(item)
 {
-  if (item.filter && !item.filter.disabled && item.filter instanceof abp.BlockingFilter)
+  if (item.filter && !item.filter.disabled && item.filter instanceof BlockingFilter)
     return null;
 
   for each (let node in item.nodes)
@@ -687,8 +634,8 @@ function compareFilter(item1, item2) {
 }
 
 function compareState(item1, item2) {
-  var state1 = (!item1.filter ? 0 : (item1.filter.disabled ? 1 : (item1.filter instanceof abp.WhitelistFilter ? 2 : 3)));
-  var state2 = (!item2.filter ? 0 : (item2.filter.disabled ? 1 : (item2.filter instanceof abp.WhitelistFilter ? 2 : 3)));
+  var state1 = (!item1.filter ? 0 : (item1.filter.disabled ? 1 : (item1.filter instanceof WhitelistFilter ? 2 : 3)));
+  var state2 = (!item2.filter ? 0 : (item2.filter.disabled ? 1 : (item2.filter instanceof WhitelistFilter ? 2 : 3)));
   return state1 - state2;
 }
 
@@ -707,6 +654,10 @@ function compareDocDomain(item1, item2)
     return -1;
   else if (item1.docDomain > item2.docDomain)
     return 1;
+  else if (item1.thirdParty && !item2.thirdParty)
+    return -1;
+  else if (!item1.thirdParty && item2.thirdParty)
+    return 1;
   else
     return 0;
 }
@@ -723,6 +674,16 @@ function createSortWithFallback(cmpFunc, fallbackFunc, desc) {
   }
 }
 
+var progressListener =
+{
+  onLocationChange: function() handleLocationChange(),
+  onProgressChange: function() {},
+  onSecurityChange: function() {},
+  onStateChange: function() {},
+  onStatusChange: function() {},
+  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebProgressListener, Ci.nsISupportsWeakReference])
+};
+
 // Item list's tree view object
 var treeView = {
   //
@@ -766,8 +727,8 @@ var treeView = {
       this.atoms[atom + "-false"] = atomService.getAtom(atom + "-false");
     }
 
-    this.itemsDummyTooltip = abp.getString("no_blocking_suggestions");
-    this.whitelistDummyTooltip = abp.getString("whitelisted_page");
+    this.itemsDummyTooltip = Utils.getString("no_blocking_suggestions");
+    this.whitelistDummyTooltip = Utils.getString("whitelisted_page");
 
     // Check current sort direction
     var cols = document.getElementsByTagName("treecol");
@@ -834,7 +795,7 @@ var treeView = {
         return (size ? size.join(" x ") : "");
       }
       else if (col == "docDomain")
-        return this.data[row].docDomain;
+        return this.data[row].docDomain + " " + (this.data[row].thirdParty ? docDomainThirdParty : docDomainFirstParty);
       else
         return this.data[row].location;
     }
@@ -844,11 +805,11 @@ var treeView = {
         return "";
 
       if (col == "filter") {
-        var filter = abp.policy.isWindowWhitelisted(window.content);
+        var filter = Policy.isWindowWhitelisted(window.content);
         return filter ? filter.text : "";
       }
 
-      return (abp.policy.isWindowWhitelisted(window.content) ? this.whitelistDummy : this.itemsDummy);
+      return (Policy.isWindowWhitelisted(window.content) ? this.whitelistDummy : this.itemsDummy);
     }
   },
 
@@ -876,11 +837,11 @@ var treeView = {
       state = "state-regular";
       if (filter && !filter.disabled)
       {
-        if (filter instanceof abp.WhitelistFilter)
+        if (filter instanceof WhitelistFilter)
           state = "state-whitelisted";
-        else if (filter instanceof abp.BlockingFilter)
+        else if (filter instanceof BlockingFilter)
           state = "state-filtered";
-        else if (filter instanceof abp.ElemHideFilter)
+        else if (filter instanceof ElemHideFilter)
           state = "state-hidden";
       }
     }
@@ -888,7 +849,7 @@ var treeView = {
       properties.AppendElement(this.atoms["dummy-true"]);
 
       state = "state-filtered";
-      if (this.data && abp.policy.isWindowWhitelisted(window.content))
+      if (this.data && Policy.isWindowWhitelisted(window.content))
         state = "state-whitelisted";
     }
     properties.AppendElement(this.atoms[state]);
@@ -968,6 +929,7 @@ var treeView = {
   filter: "",
   data: null,
   allData: [],
+  dataMap: {__proto__: null},
   sortColumn: null,
   sortProc: null,
   resortTimeout: null,
@@ -975,6 +937,7 @@ var treeView = {
   whitelistDummy: null,
   itemsDummyTooltip: null,
   whitelistDummyTooltip: null,
+  itemToSelect: null,
 
   sortProcs: {
     address: sortByAddress,
@@ -991,36 +954,47 @@ var treeView = {
     docDomainDesc: createSortWithFallback(compareDocDomain, sortByAddress, true)
   },
 
-  setData: function(data) {
+  clearData: function(data) {
     var oldRows = this.rowCount;
 
-    this.allData = data;
-    for each (let item in this.allData)
-    {
-      if (item.filter instanceof abp.RegExpFilter && item.filter.disabled)
-        item.filter = null;
-      if (!item.filter)
-        item.filter = disabledWhitelistMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
-      if (!item.filter)
-        item.filter = disabledBlacklistMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
-    }
+    this.allData = [];
+    this.dataMap = {__proto__: null};
     this.refilter();
 
     this.boxObject.rowCountChanged(0, -oldRows);
     this.boxObject.rowCountChanged(0, this.rowCount);
   },
 
-  addItem: function(item) {
+  addItem: function(/**Node*/ node, /**RequestEntry*/ item, /**Boolean*/ scanComplete)
+  {
+    // Merge duplicate entries
+    let key = item.location + " " + item.type + " " + item.docDomain;
+    if (key in this.dataMap)
+    {
+      // We know this item already - take over the filter if any and be done with it
+      let existing = this.dataMap[key];
+      if (item.filter)
+        existing.filter = item.filter;
+
+      existing.nodes.push(node);
+      this.invalidateItem(existing);
+      return;
+    }
+
+    // Add new item to the list
+    // Store original item in orig property - reading out prototype is messed up in Gecko 1.9.2
+    item = {__proto__: item, orig: item, nodes: [node]};
     this.allData.push(item);
+    this.dataMap[key] = item;
+
+    // Show disabled filters if no other filter applies
     if (!item.filter)
-      item.filter = disabledWhitelistMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
-    if (!item.filter)
-      item.filter = disabledBlacklistMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
+      item.filter = disabledMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
 
     if (!this.matchesFilter(item))
       return;
 
-    var index = -1;
+    let index = -1;
     if (this.sortProc && this.sortColumn && this.sortColumn.id == "size")
     {
       // Sorting by size requires accessing content document, and that's
@@ -1052,6 +1026,27 @@ var treeView = {
       this.boxObject.invalidateRow(0);
     else
       this.boxObject.rowCountChanged(index, 1);
+
+    if (this.itemToSelect == key)
+    {
+      this.selection.select(index);
+      this.boxObject.ensureRowIsVisible(index);
+      this.itemToSelect = null;
+    }
+    else if (!scanComplete && this.selection.currentIndex >= 0) // Keep selected row visible while scanning
+      this.boxObject.ensureRowIsVisible(this.selection.currentIndex);
+  },
+
+  updateFilters: function()
+  {
+    for each (let item in this.allData)
+    {
+      if (item.filter instanceof RegExpFilter && item.filter.disabled)
+        delete item.filter;
+      if (!item.filter)
+        item.filter = disabledMatcher.matchesAny(item.location, item.typeDescr, item.docDomain, item.thirdParty);
+    }
+    this.refilter();
   },
 
   /**
@@ -1079,7 +1074,10 @@ var treeView = {
 
     return (item.location.toLowerCase().indexOf(this.filter) >= 0 ||
             (item.filter && item.filter.text.toLowerCase().indexOf(this.filter) >= 0) ||
-            item.localizedDescr.toLowerCase().indexOf(this.filter) >= 0);
+            item.localizedDescr.toLowerCase().indexOf(this.filter) >= 0 ||
+            (item.docDomain && item.docDomain.toLowerCase().indexOf(this.filter) >= 0) ||
+            (item.docDomain && item.thirdParty && docDomainThirdParty.toLowerCase().indexOf(this.filter) >= 0) ||
+            (item.docDomain && !item.thirdParty && docDomainFirstParty.toLowerCase().indexOf(this.filter) >= 0));
   },
 
   setFilter: function(filter) {
@@ -1152,23 +1150,17 @@ var treeView = {
     if (!this.data || this.data.length)
       return null;
 
-    var filter = abp.policy.isWindowWhitelisted(window.content);
+    var filter = Policy.isWindowWhitelisted(window.content);
     if (filter)
       return {tooltip: this.whitelistDummyTooltip, filter: filter};
     else
       return {tooltip: this.itemsDummyTooltip};
   },
 
-  selectItem: function(item) {
-    var row = -1;
-    for (var i = 0; row < 0 && i < this.data.length; i++)
-      if (this.data[i] == item)
-        row = i;
-
-    if (row < 0 )
-      return;
-
-    this.selection.select(row);
-    this.boxObject.ensureRowIsVisible(row);
+  invalidateItem: function(item)
+  {
+    let row = this.data.indexOf(item);
+    if (row >= 0)
+      this.boxObject.invalidateRow(row);
   }
 }
