@@ -15,7 +15,7 @@
  *
  * The Initial Developer of the Original Code is
  * Wladimir Palant.
- * Portions created by the Initial Developer are Copyright (C) 2006-2010
+ * Portions created by the Initial Developer are Copyright (C) 2006-2011
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s):
@@ -23,6 +23,7 @@
  * ***** END LICENSE BLOCK ***** */
 
 let newInstall = true;
+let savedDialogFlex = null;
 let editMode = true;
 let autoAdd = false;
 let source = null;
@@ -37,27 +38,73 @@ let suppressResize = true;
 
 let closing = false;
 let subscriptionListLoading = false;
-let appLocale = "en-US";
 let otherButton = null;
 
-let adblockID = "{34274bf4-1d97-a289-e984-17e546307e4f}";
-let filtersetG = "filtersetg@updater";
-
 function init()
 {
   if (window.arguments  && window.arguments.length)
   {
-    // In K-Meleon we might get the arguments wrapped
-    for (var i = 0; i < window.arguments.length; i++)
-      if (window.arguments[i] && "wrappedJSObject" in window.arguments[i])
-        window.arguments[i] = window.arguments[i].wrappedJSObject;
-
     newInstall = false;
     [source, result] = window.arguments;
     if (window.arguments.length > 2 && window.arguments[2])
       window.hasSubscription = window.arguments[2];
   }
 
+  if (newInstall && Utils.isFennec)
+  {
+    // HACK: In Fennec 4.0 menulist elements won't work "by themselves". We
+    // have to go to the top level and trigger MenuListHelperUI manually.
+    let topWnd = window.QueryInterface(Ci.nsIInterfaceRequestor)
+                       .getInterface(Ci.nsIWebNavigation)
+                       .QueryInterface(Ci.nsIDocShellTreeItem)
+                       .rootTreeItem
+                       .QueryInterface(Ci.nsIInterfaceRequestor)
+                       .getInterface(Ci.nsIDOMWindow);
+    if (topWnd.wrappedJSObject)
+      topWnd = topWnd.wrappedJSObject;
+
+    let menulist = E("subscriptions");
+    if ("MenuListHelperUI" in topWnd && menulist.parentNode.localName != "stack")
+    {
+      // Add a layer on top of the menulist to handle clicks, menulist clicks
+      // are otherwise ignored and cannot be intercepted
+      let stack = document.createElement("stack");
+      menulist.parentNode.replaceChild(stack, menulist);
+      stack.appendChild(menulist);
+
+      let clickLayer = document.createElement("hbox");
+      stack.appendChild(clickLayer);
+
+      clickLayer.addEventListener("click", function(event)
+      {
+        if (event.button == 0 && !menulist.disabled && menulist.itemCount)
+        {
+          menulist.focus();
+          topWnd.MenuListHelperUI.show(menulist);
+        }
+      }, true);
+
+      // menulist needs to be initialized after being moved, re-run init() later
+      Utils.runAsync(init);
+      return;
+    }
+
+    // The template is being displayed as a list item, remove it
+    let subscriptionsTemplate = E("subscriptionsTemplate");
+    if (subscriptionsTemplate && subscriptionsTemplate.parentNode)
+      subscriptionsTemplate.parentNode.removeChild(subscriptionsTemplate);
+
+    // window.close() closes the entire window (bug 642604), make sure to close
+    // only a single tab instead.
+    if ("BrowserUI" in topWnd)
+    {
+      window.close = function()
+      {
+        topWnd.BrowserUI.closeTab();
+      };
+    }
+  }
+
   if (!result)
   {
     result = {};
@@ -74,6 +121,20 @@ function init()
       source.mainSubscriptionURL = source.mainSubscriptionTitle = null;
   }
 
+  if (newInstall && !Utils.isFennec)
+  {
+    // HACK: We will remove dialog content box flex if a subscription is
+    // selected, need to find the content box and save it flex value.
+    let docContent = document.getAnonymousNodes(document.documentElement);
+    docContent = (docContent && docContent.length ? docContent[0] : null);
+    if (docContent && docContent.hasAttribute("class") &&
+        /\bdialog-content-box\b/.test(docContent.getAttribute("class")) &&
+        docContent.hasAttribute("flex"))
+    {
+      savedDialogFlex = [docContent, docContent.getAttribute("flex")]
+    }
+  }
+
   E("description-newInstall").hidden = !newInstall;
   if (newInstall)
     document.documentElement.setAttribute("newInstall", "true");
@@ -81,9 +142,9 @@ function init()
   E("subscriptionsBox").hidden = E("all-subscriptions-container").hidden
     = E("subscriptionInfo").hidden = editMode;
 
-  E("fromWebText").hidden = !editMode || source instanceof abp.Subscription;
-  E("editText").hidden = !(source instanceof abp.Subscription) || source instanceof abp.ExternalSubscription;
-  E("externalText").hidden = !(source instanceof abp.ExternalSubscription);
+  E("fromWebText").hidden = !editMode || source instanceof Subscription;
+  E("editText").hidden = !(source instanceof Subscription) || source instanceof ExternalSubscription;
+  E("externalText").hidden = !(source instanceof ExternalSubscription);
   E("differentSubscription").hidden = !editMode;
 
   otherButton = document.documentElement.getButton("extra2");
@@ -95,7 +156,9 @@ function init()
     link.setAttribute("class", "text-link");
     link.setAttribute("value", otherButton.label);
     link.setAttribute("accesskey", otherButton.accessKey);
-    link.setAttribute("control", "otherButton")
+    link.setAttribute("control", "otherButton");
+    link.setAttribute("flex", "1");
+    link.setAttribute("crop", "end");
 
     let handler = new Function("event", document.documentElement.getAttribute("ondialogextra2"));
     link.addEventListener("command", handler, false);
@@ -118,13 +181,13 @@ function init()
   setCustomSubscription(source.title, source.url,
                         source.mainSubscriptionTitle, source.mainSubscriptionURL);
 
-  if (source instanceof abp.Subscription)
+  if (source instanceof Subscription)
   {
     document.title = document.documentElement.getAttribute("edittitle");
     document.documentElement.getButton("accept").setAttribute("label", document.documentElement.getAttribute("buttonlabelacceptedit"))
   }
 
-  if (source instanceof abp.ExternalSubscription)
+  if (source instanceof ExternalSubscription)
   {
     E("location").setAttribute("disabled", "true");
     E("autoDownload").setAttribute("disabled", "true");
@@ -133,16 +196,6 @@ function init()
   else
     E("autoDownload").checked = source.autoDownload;
 
-  // Find filter subscription suggestion based on user's browser locale
-  try
-  {
-    appLocale = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(Ci.nsIXULChromeRegistry).getSelectedLocale("adblockplus");
-  }
-  catch (e)
-  {
-    Cu.reportError(e);
-  }
-
   initialized = true;
 
   if (!editMode)
@@ -162,7 +215,7 @@ function init()
       if (!selectedItem)
         selectedItem = item;
   
-      let prefix = checkPrefixMatch(prefixes, appLocale);
+      let prefix = checkPrefixMatch(prefixes, Utils.appLocale);
       if (prefix)
       {
         item.setAttribute("class", "localeMatch");
@@ -189,19 +242,6 @@ function init()
       }
     }
     list.selectedItem = selectedItem;
-
-    // Warn if Adblock or Filterset.G Updater are installed
-    if (isExtensionActive(adblockID))
-      E("adblock-warning").hidden = false;
-  
-    if (isExtensionActive(filtersetG))
-      E("filtersetg-warning").hidden = false;
-  
-    if ("Filterset.G" in filterStorage.knownSubscriptions &&
-        !filterStorage.knownSubscriptions["Filterset.G"].disabled)
-    {
-      E("filtersetg-warning").hidden = false;
-    }
   }
 
   // Only resize if we are a chrome window (not loaded into a browser tab)
@@ -282,6 +322,15 @@ function onSelectionChange()
     E("title").focus();
   }
 
+  if (savedDialogFlex)
+  {
+    let [docContent, flex] = savedDialogFlex;
+    if (selectedSubscription)
+      docContent.removeAttribute("flex");
+    else
+      docContent.setAttribute("flex", flex);
+  }
+
   updateSubscriptionInfo();
 }
 
@@ -323,11 +372,13 @@ function loadSubscriptionList()
     return;
 
   E("all-subscriptions-container").selectedIndex = 0;
+  E("all-subscriptions-loading").hidden = false;
 
   let request = new XMLHttpRequest();
   let errorHandler = function()
   {
     E("all-subscriptions-container").selectedIndex = 2;
+    E("all-subscriptions-loading").hidden = true;
   };
   let successHandler = function()
   {
@@ -348,7 +399,7 @@ function loadSubscriptionList()
     }
   };
 
-  request.open("GET", abp.prefs.subscriptions_listurl);
+  request.open("GET", Prefs.subscriptions_listurl);
   request.onerror = errorHandler;
   request.onload = successHandler;
   request.send(null);
@@ -364,6 +415,7 @@ function processSubscriptionList(doc)
 
   addSubscriptions(list, doc.documentElement, 0, null, null);
   E("all-subscriptions-container").selectedIndex = 1;
+  E("all-subscriptions-loading").hidden = true;
 }
 
 function addSubscriptions(list, parent, level, parentTitle, parentURL)
@@ -405,10 +457,10 @@ function addSubscriptions(list, parent, level, parentTitle, parentURL)
       let title = document.createElement("description");
       if (isFirst)
       {
-        if (checkPrefixMatch(node.getAttribute("prefixes"), appLocale))
-          title.setAttribute("class", "title localeMatch");
+        if (checkPrefixMatch(node.getAttribute("prefixes"), Utils.appLocale))
+          title.setAttribute("class", "subscriptionTitle localeMatch");
         else
-          title.setAttribute("class", "title");
+          title.setAttribute("class", "subscriptionTitle");
         title.textContent = node.getAttribute("title");
         mainTitle = variant.getAttribute("title");
         mainURL = variant.getAttribute("url");
@@ -469,7 +521,7 @@ function setCustomSubscription(title, url, mainSubscriptionTitle, mainSubscripti
     let link = document.createElement("label");
     link.className = "text-link";
     link.setAttribute("tooltiptext", mainSubscriptionURL);
-    link.addEventListener("click", function() abp.loadInBrowser(mainSubscriptionURL), false);
+    link.addEventListener("click", function() Utils.loadInBrowser(mainSubscriptionURL), false);
     link.textContent = mainSubscriptionTitle;
     messageElement.appendChild(link);
     messageElement.appendChild(document.createTextNode(afterLink));
@@ -501,7 +553,7 @@ function validateURL(url)
   } catch (e) {}
 
   // Is this a valid URL?
-  let uri = abp.makeURL(url);
+  let uri = Utils.makeURI(url);
   if (uri)
     return uri.spec;
 
@@ -523,11 +575,11 @@ function addSubscription()
   else
   {
     url = E("location").value;
-    if (!(source instanceof abp.ExternalSubscription))
+    if (!(source instanceof ExternalSubscription))
       url = validateURL(url);
     if (!url)
     {
-      abp.alert(window, abp.getString("subscription_invalid_location"));
+      Utils.alert(window, Utils.getString("subscription_invalid_location"));
       E("location").focus();
       return false;
     }
@@ -553,24 +605,56 @@ function addSubscription()
 
   if (autoAdd)
   {
-    abp.addSubscription(result.url, result.title, result.autoDownload, result.disabled);
+    doAddSubscription(result.url, result.title, result.autoDownload, result.disabled);
     if ("mainSubscriptionURL" in result)
-      abp.addSubscription(result.mainSubscriptionURL, result.mainSubscriptionTitle, result.autoDownload, result.disabled);
+      doAddSubscription(result.mainSubscriptionURL, result.mainSubscriptionTitle, result.autoDownload, result.disabled);
   }
 
   closing = true;
   return true;
 }
 
+/**
+ * Adds a new subscription to the list.
+ */
+function doAddSubscription(/**String*/ url, /**String*/ title, /**Boolean*/ autoDownload, /**Boolean*/ disabled)
+{
+  if (typeof autoDownload == "undefined")
+    autoDownload = true;
+  if (typeof disabled == "undefined")
+    disabled = false;
+
+  let subscription = Subscription.fromURL(url);
+  if (!subscription)
+    return;
+
+  FilterStorage.addSubscription(subscription);
+
+  if (disabled != subscription.disabled)
+  {
+    subscription.disabled = disabled;
+    FilterStorage.triggerObservers(disabled ? "subscriptions disable" : "subscriptions enable", [subscription]);
+  }
+
+  subscription.title = title;
+  if (subscription instanceof DownloadableSubscription)
+    subscription.autoDownload = autoDownload;
+  FilterStorage.triggerObservers("subscriptions updateinfo", [subscription]);
+
+  if (subscription instanceof DownloadableSubscription && !subscription.lastDownload)
+    Synchronizer.execute(subscription);
+  FilterStorage.saveToDisk();
+}
+
 function hasSubscription(url)
 {
-  return abp.filterStorage.subscriptions.some(function(subscription) subscription instanceof abp.DownloadableSubscription && subscription.url == url);
+  return FilterStorage.subscriptions.some(function(subscription) subscription instanceof DownloadableSubscription && subscription.url == url);
 }
 
 function checkUnload()
 {
   if (newInstall && !closing)
-    return abp.getString("subscription_notAdded_warning");
+    return Utils.getString("subscription_notAdded_warning");
 
   return undefined;
 }
@@ -581,75 +665,7 @@ function onDialogCancel()
   if (!message)
     return true;
 
-  message += " " + abp.getString("subscription_notAdded_warning_addendum");
-  closing = abp.confirm(window, message);
+  message += " " + Utils.getString("subscription_notAdded_warning_addendum");
+  closing = Utils.confirm(window, message);
   return closing;
 }
-
-function uninstallExtension(id)
-{
-  let extensionManager = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
-  if (extensionManager.getItemForID(id))
-  {
-    let location = extensionManager.getInstallLocation(id);
-    if (location && !location.canAccess)
-    {
-      // Cannot uninstall, need to disable
-      extensionManager.disableItem(id);
-    }
-    else
-    {
-      extensionManager.uninstallItem(id);
-    }
-  }
-}
-
-function isExtensionActive(id)
-{
-  if (!("@mozilla.org/extensions/manager;1" in Cc))
-    return false;
-
-  let extensionManager = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
-
-  // First check whether the extension is installed
-  if (!extensionManager.getItemForID(id))
-    return false;
-
-  let ds = extensionManager.datasource;
-  let rdfService = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);
-  let source = rdfService.GetResource("urn:mozilla:item:" + id);
-
-  // Check whether extension is disabled
-  let link = rdfService.GetResource("http://www.mozilla.org/2004/em-rdf#isDisabled");
-  let target = ds.GetTarget(source, link, true);
-  if (target instanceof Ci.nsIRDFLiteral && target.Value == "true")
-    return false;
-
-  // Check whether an operation is pending for the extension
-  link = rdfService.GetResource("http://www.mozilla.org/2004/em-rdf#opType");
-  if (ds.GetTarget(source, link, false))
-    return false;
-
-  return true;
-}
-
-function uninstallAdblock()
-{
-  uninstallExtension(adblockID);
-  E("adblock-warning").hidden = true;
-}
-
-function uninstallFiltersetG()
-{
-  // Disable further updates
-  abp.denyFiltersetG = true;
-
-  // Uninstall extension
-  uninstallExtension(filtersetG);
-
-  // Remove filter subscription
-  if ("Filterset.G" in filterStorage.knownSubscriptions)
-    filterStorage.removeSubscription(filterStorage.knownSubscriptions["Filterset.G"]);
-
-  E("filtersetg-warning").hidden = true;
-}
